import uuid

from .signals import post_version_operation, pre_version_operation


def send_pre_version_operation(operation, version, **kwargs):
    """
    Signal emitter for before a version operation occurs.
    A token is emitted that will allow the pre and post emitted signals to be tied together.

    :param operation: Operation constants
    :param version: Version instance
    :param kwargs:
    :return: A unique token for the transaction
    """
    token = str(uuid.uuid4())
    pre_version_operation.send(
        sender=version.content_type.model_class(),
        operation=operation,
        token=token,
        obj=version,
        **kwargs
    )
    return token


def send_post_version_operation(operation, version, token, **kwargs):
    """
    Signal emitter for after a version operation occurs

    :param operation: Operation constants
    :param version: Version instance
    :param token: A unique token for the transaction
    :param kwargs:
    """
    post_version_operation.send(
        sender=version.content_type.model_class(),
        operation=operation,
        token=token,
        obj=version,
        **kwargs
    )



from cms.admin.utils import GrouperModelAdmin
from django.contrib import admin
from django.urls import re_path

from djangocms_versioning.admin import (
    ExtendedGrouperVersionAdminMixin,
    ExtendedVersionAdminMixin,
)

from .models import Answer, Poll, PollContent
from .views import PreviewView


@admin.register(PollContent)
class PollContentAdmin(ExtendedVersionAdminMixin, admin.ModelAdmin):
    list_display = ("text", )

    def get_urls(self):
        info = self.model._meta.app_label, self.model._meta.model_name
        return [
            re_path(
                r"^(?P<id>\d+)/preview/$",
                self.admin_site.admin_view(PreviewView.as_view()),
                name="{}_{}_preview".format(*info),
            )
        ] + super().get_urls()


@admin.register(Poll)
class PollAdmin(ExtendedGrouperVersionAdminMixin, GrouperModelAdmin):
    list_display = ("content__text", "get_author", "get_modified_date", "get_versioning_state", "admin_list_actions")


@admin.register(Answer)
class AnswerAdmin(admin.ModelAdmin):
    pass



import copy

from cms.models import PageContent
from cms.test_utils.testcases import CMSTestCase
from django.apps import apps

from djangocms_versioning.constants import ARCHIVED, PUBLISHED
from djangocms_versioning.datastructures import VersionableItem, default_copy
from djangocms_versioning.models import Version
from djangocms_versioning.test_utils.factories import PollVersionFactory
from djangocms_versioning.test_utils.people.models import PersonContent
from djangocms_versioning.test_utils.polls.models import Poll, PollContent


class VersionableItemTestCase(CMSTestCase):
    def setUp(self):
        self.initial_version = PollVersionFactory()

    def test_distinct_groupers(self):
        latest_poll1_version = PollVersionFactory(
            content__poll=self.initial_version.content.poll
        )
        poll2_version = PollVersionFactory()
        PollVersionFactory(content__poll=poll2_version.content.poll)
        latest_poll2_version = PollVersionFactory(
            content__poll=poll2_version.content.poll
        )

        versionable = VersionableItem(
            content_model=PollContent,
            grouper_field_name="poll",
            copy_function=default_copy,
        )
        self.assertQuerysetEqual(
            versionable.distinct_groupers(),
            [latest_poll1_version.content.pk, latest_poll2_version.content.pk],
            transform=lambda x: x.pk,
            ordered=False,
        )

    def test_queryset_filter_for_distinct_groupers(self):
        poll1_archived_version = PollVersionFactory(
            content__poll=self.initial_version.content.poll, state=ARCHIVED
        )
        poll1_published_version = PollVersionFactory(
            content__poll=self.initial_version.content.poll, state=PUBLISHED
        )
        poll2_version = PollVersionFactory()
        PollVersionFactory(content__poll=poll2_version.content.poll, state=ARCHIVED)
        poll2_archived_version = PollVersionFactory(
            content__poll=poll2_version.content.poll, state=ARCHIVED
        )

        versionable = VersionableItem(
            content_model=PollContent,
            grouper_field_name="poll",
            copy_function=default_copy,
        )

        qs_published_filter = {"versions__state__in": [PUBLISHED]}
        # Should be one published version
        self.assertQuerysetEqual(
            versionable.distinct_groupers(**qs_published_filter),
            [poll1_published_version.content.pk],
            transform=lambda x: x.pk,
            ordered=False,
        )

        qs_archive_filter = {"versions__state__in": [ARCHIVED]}
        # Should be two archived versions
        self.assertQuerysetEqual(
            versionable.distinct_groupers(**qs_archive_filter),
            [poll1_archived_version.content.pk, poll2_archived_version.content.pk],
            transform=lambda x: x.pk,
            ordered=False,
        )

    def test_for_grouper(self):
        poll1_version2 = PollVersionFactory(
            content__poll=self.initial_version.content.poll
        )
        poll2_version = PollVersionFactory()
        PollVersionFactory(content__poll=poll2_version.content.poll)
        PollVersionFactory(content__poll=poll2_version.content.poll)

        versionable = VersionableItem(
            content_model=PollContent,
            grouper_field_name="poll",
            copy_function=default_copy,
        )

        self.assertQuerysetEqual(
            versionable.for_grouper(self.initial_version.content.poll),
            [self.initial_version.content.pk, poll1_version2.content.pk],
            transform=lambda x: x.pk,
            ordered=False,
        )

    def test_grouper_model(self):
        versionable = VersionableItem(
            content_model=PollContent,
            grouper_field_name="poll",
            copy_function=default_copy,
        )

        self.assertEqual(versionable.grouper_model, Poll)

    def test_content_model_is_sideframe_editable_for_sideframe_disabled_model(self):
        """
        A content model with placeholders should not be opened in the sideframe
        """
        versionable = VersionableItem(
            content_model=PageContent,
            grouper_field_name="page",
            copy_function=default_copy,
        )

        self.assertEqual(versionable.content_model_is_sideframe_editable, False)

    def test_content_model_is_sideframe_editable_for_sideframe_enabled_model(self):
        """
        A content model without placeholders should be opened in the sideframe
        """
        versionable = VersionableItem(
            content_model=PollContent,
            grouper_field_name="poll",
            copy_function=default_copy,
        )

        self.assertEqual(versionable.content_model_is_sideframe_editable, True)


class VersionableItemProxyModelTestCase(CMSTestCase):
    @classmethod
    def setUpClass(cls):
        cls._all_models = copy.deepcopy(apps.all_models)

    @classmethod
    def tearDownClass(cls):
        apps.all_models = cls._all_models

    def tearDown(self):
        apps.all_models.pop("djangocms_versioning", None)

    def test_version_model_proxy(self):
        versionable = VersionableItem(
            content_model=PersonContent,
            grouper_field_name="person",
            copy_function=default_copy,
        )
        version_model_proxy = versionable.version_model_proxy

        self.assertIn(Version, version_model_proxy.mro())
        self.assertEqual(version_model_proxy.__name__, "PersonContentVersion")
        self.assertEqual(version_model_proxy._source_model, PersonContent)
        self.assertTrue(version_model_proxy._meta.proxy)

    def test_version_model_proxy_cached(self):
        """Test that version_model_proxy property is cached
        and return value is created once."""
        versionable = VersionableItem(
            content_model=PersonContent,
            grouper_field_name="person",
            copy_function=default_copy,
        )

        self.assertEqual(
            id(versionable.version_model_proxy), id(versionable.version_model_proxy)
        )



from cms.toolbar.items import ButtonList
from cms.toolbar.toolbar import CMSToolbar
from django.test import RequestFactory

from djangocms_versioning.cms_toolbars import VersioningToolbar
from djangocms_versioning.test_utils.factories import UserFactory


def get_toolbar(content_obj, user=None, **kwargs):
    """
    Helper method to set up the toolbar

    Warning: Other apps may use this helper to improve readability and stability of tests.
            Be sure to keep backwards compatibility where possible
    """
    # Set the user if none are sent
    if not user:
        user = UserFactory(is_staff=True)

    request = kwargs.get("request", RequestFactory().get("/"))
    request.user = user
    request.session = kwargs.get("session", {})
    request.current_page = getattr(content_obj, "page", None)
    request.toolbar = CMSToolbar(request)
    # Set the toolbar class
    if kwargs.get("toolbar_class", False):
        toolbar_class = kwargs.get("toolbar_class")
    else:
        toolbar_class = VersioningToolbar
    toolbar = toolbar_class(
        request, toolbar=request.toolbar, is_current_app=True, app_path="/"
    )
    toolbar.toolbar.set_object(content_obj)
    # Set the toolbar mode
    if kwargs.get("edit_mode", False):
        toolbar.toolbar.edit_mode_active = True
        toolbar.toolbar.preview_mode_active = False
        toolbar.toolbar.structure_mode_active = False
    elif kwargs.get("preview_mode", False):
        toolbar.toolbar.edit_mode_active = False
        toolbar.toolbar.preview_mode_active = True
        toolbar.toolbar.structure_mode_active = False
    elif kwargs.get("structure_mode", False):
        toolbar.toolbar.edit_mode_active = False
        toolbar.toolbar.preview_mode_active = False
        toolbar.toolbar.structure_mode_active = True
    toolbar.populate()
    return toolbar


def find_toolbar_buttons(button_name, toolbar):
    """
    Returns a button that matches the button name

    Warning: Other apps may use this helper to improve readability and stability of tests.
        Be sure to keep backwards compatibility where possible
    """
    found = []
    for button_list in toolbar.get_right_items():
        if isinstance(button_list, ButtonList):
            found = found + [
                button for button in button_list.buttons if button.name == button_name
            ]
    return found


def toolbar_button_exists(button_name, toolbar):
    """
    Check to see if a button exists in a supplied toolbar

    Warning: Other apps may use this helper to improve readability and stability of tests.
        Be sure to keep backwards compatibility where possible
    """
    found = find_toolbar_buttons(button_name, toolbar)
    return bool(len(found))



from django.db import models


class IncorrectBlogPost(models.Model):
    name = models.TextField()

    def __str__(self):
        return f"{self.name} ({self.pk})"


class IncorrectBlogContent(models.Model):
    blogpost = models.ForeignKey(IncorrectBlogPost, on_delete=models.CASCADE)
    text = models.TextField()

    def __str__(self):
        return self.text



# Generated by Django 4.2 on 2023-05-14 10:27

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('djangocms_versioning', '0016_alter_version_content_type'),
        ('djangocms_versioning', '0016_auto_20230505_0934'),
    ]

    operations = [
    ]



from cms.app_base import CMSAppConfig

from djangocms_versioning.datastructures import VersionableItem, default_copy

from .models import IncorrectBlogContent


def blog_method(obj, field):
    return getattr(obj, field)


class IncorrectBlogpostCMSConfig(CMSAppConfig):
    djangocms_versioning_enabled = True
    versioning = [
        VersionableItem(
            content_model=IncorrectBlogContent,
            grouper_field_name="blogpost",
            copy_function=default_copy,
        ),
    ]
    extended_admin_field_modifiers = [
        {IncorrectBlogContent: {"non_existent_field": blog_method}},
    ]



from django.utils.translation import gettext_lazy as _

"""Version states"""
ARCHIVED = "archived"
DRAFT = "draft"
PUBLISHED = "published"
UNPUBLISHED = "unpublished"

VERSION_STATES = (
    (DRAFT, _("Draft")),
    (PUBLISHED, _("Published")),
    (UNPUBLISHED, _("Unpublished")),
    (ARCHIVED, _("Archived")),
)
"""Version operation states"""
OPERATION_ARCHIVE = "operation_archive"
OPERATION_DRAFT = "operation_draft"
OPERATION_PUBLISH = "operation_publish"
OPERATION_UNPUBLISH = "operation_unpublish"

INDICATOR_DESCRIPTIONS = {
    "published": _("Published"),
    "dirty": _("Changed"),
    "draft": _("Draft"),
    "unpublished": _("Unpublished"),
    "archived": _("Archived"),
    "empty": _("Empty"),
}



from cms.utils.urlutils import admin_reverse
from django.contrib.auth import get_permission_codename
from django.utils.http import urlencode
from django.utils.translation import gettext_lazy as _

from .constants import ARCHIVED, DRAFT, PUBLISHED, UNPUBLISHED, VERSION_STATES
from .models import Version


def _reverse_action(version, action, back=None):
    get_params = f"?{urlencode({'back': back})}" if back else ""
    return admin_reverse(
        f"{version._meta.app_label}_{version.versionable.version_model_proxy._meta.model_name}_{action}",
        args=(version.pk,)
    ) + get_params


def content_indicator_menu(request, status, versions, back=""):
    from djangocms_versioning.helpers import version_list_url

    menu = []
    if request.user.has_perm(f"cms.{get_permission_codename('change', versions[0]._meta)}"):
        if versions[0].check_unlock.as_bool(request.user):
            can_unlock = request.user.has_perm("djangocms_versioning.delete_versionlock")
            # disable if permissions are insufficient
            additional_class = "" if can_unlock else " cms-pagetree-dropdown-item-disabled"
            menu.append((
                _("Unlock (%(message)s)") % {"message": versions[0].locked_message()}, "cms-icon-unlock",
                _reverse_action(versions[0], "unlock"),
                "js-cms-tree-lang-trigger" + additional_class,  # Triggers POST from the frontend
            ))
        if versions[0].check_publish.as_bool(request.user):
            menu.append((
                _("Publish"), "cms-icon-publish",
                _reverse_action(versions[0], "publish"),
                "js-cms-tree-lang-trigger",  # Triggers POST from the frontend
            ))
        if versions[0].check_edit_redirect.as_bool(request.user) and versions[0].state == PUBLISHED:
            menu.append((
                _("Create new draft"), "cms-icon-edit-new",
                _reverse_action(versions[0], "edit_redirect"),
                "js-cms-tree-lang-trigger js-cms-pagetree-page-view",  # Triggers POST from the frontend
            ))
        if versions[0].check_revert.as_bool(request.user) and versions[0].state == UNPUBLISHED:
            # Do not offer revert from unpublish -> archived versions to be managed in version admin
            label = _("Revert from Unpublish")
            menu.append((
                label, "cms-icon-undo",
                _reverse_action(versions[0], "revert"),
                "js-cms-tree-lang-trigger",  # Triggers POST from the frontend
            ))
        if versions[0].check_unpublish.as_bool(request.user):
            menu.append((
                _("Unpublish"), "cms-icon-unpublish",
                _reverse_action(versions[0], "unpublish"),
                "js-cms-tree-lang-trigger",
            ))
        if len(versions) > 1 and versions[1].check_unpublish.as_bool(request.user):
            menu.append((
                _("Unpublish"), "cms-icon-unpublish",
                _reverse_action(versions[1], "unpublish"),
                "js-cms-tree-lang-trigger",
            ))
        if versions[0].check_discard.as_bool(request.user):
            menu.append((
                _("Delete Draft") if status == DRAFT else _("Discard Changes"), "cms-icon-bin",
                _reverse_action(versions[0], "discard", back=back),
                "",  # Let view ask for confirmation
            ))
        if len(versions) >= 2 and versions[0].state == DRAFT and versions[1].state == PUBLISHED:
            menu.append((
                _("Compare Draft to Published..."), "cms-icon-layers",
                _reverse_action(versions[1], "compare") +
                "?" + urlencode({
                    "compare_to": versions[0].pk,
                    "back": back,
                }),
                "",
            ))
    menu.append(
        (
            _("Manage Versions..."), "cms-icon-copy",
            version_list_url(versions[0].content),
            "",
        )
    )
    return menu


def content_indicator(content_obj):
    """Translates available versions into status to be reflected by the indicator.
    Function caches the result in the page_content object"""

    if not content_obj:
        return None  # pragma: no cover
    elif not hasattr(content_obj, "_indicator_status"):
        versions = Version.objects.filter_by_content_grouping_values(
            content_obj
        ).order_by("-pk")
        signature = {
            state: versions.filter(state=state)
            for state, name in VERSION_STATES
        }
        if signature[DRAFT] and not signature[PUBLISHED]:
            content_obj._indicator_status = "draft"
            content_obj._version = signature[DRAFT]
        elif signature[DRAFT] and signature[PUBLISHED]:
            content_obj._indicator_status = "dirty"
            content_obj._version = (signature[DRAFT][0], signature[PUBLISHED][0])
        elif signature[PUBLISHED]:
            content_obj._indicator_status = "published"
            content_obj._version = signature[PUBLISHED]
        elif versions[0].state == UNPUBLISHED:
            content_obj._indicator_status = "unpublished"
            content_obj._version = signature[UNPUBLISHED]
        elif versions[0].state == ARCHIVED:
            content_obj._indicator_status = "archived"
            content_obj._version = signature[ARCHIVED]
        else:  # pragma: no cover
            content_obj._indicator_status = None
            content_obj._version = [None]
    return content_obj._indicator_status


def is_editable(content_obj, request):
    """Check of content_obj is editable"""
    if not content_obj.content_indicator():
        # Something's wrong: content indicator not identified. Maybe no version?
        return False
    versions = content_obj._version
    return versions[0].check_modify.as_bool(request.user)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.14 on 2018-08-09 16:14
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("contenttypes", "0002_remove_content_type_name"),
        ("djangocms_versioning", "0005_remove_version_label"),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name="version", unique_together=set([("content_type", "object_id")])
        )
    ]



from collections import OrderedDict
from copy import copy

from cms.cms_toolbars import (
    ADD_PAGE_LANGUAGE_BREAK,
    LANGUAGE_MENU_IDENTIFIER,
    PageToolbar,
    PlaceholderToolbar,
)
from cms.models import PageContent
from cms.toolbar.items import RIGHT, Break, ButtonList, TemplateItem
from cms.toolbar.utils import get_object_preview_url
from cms.toolbar_pool import toolbar_pool
from cms.utils import page_permissions
from cms.utils.conf import get_cms_setting
from cms.utils.i18n import get_language_dict, get_language_tuple
from cms.utils.urlutils import add_url_parameters, admin_reverse
from django.apps import apps
from django.conf import settings
from django.contrib.auth import get_permission_codename
from django.contrib.contenttypes.models import ContentType
from django.urls import reverse
from django.utils.http import urlencode
from django.utils.translation import gettext_lazy as _

from djangocms_versioning.conf import LOCK_VERSIONS
from djangocms_versioning.constants import DRAFT, PUBLISHED
from djangocms_versioning.helpers import (
    get_latest_admin_viewable_content,
    version_list_url,
)
from djangocms_versioning.models import Version

VERSIONING_MENU_IDENTIFIER = "version"


class VersioningToolbar(PlaceholderToolbar):
    class Media:
        js = ("cms/js/admin/actions.js",)

    def _get_versionable(self):
        """Helper method to get the versionable for the content type
        of the version
        """
        versioning_extension = apps.get_app_config("djangocms_versioning").cms_extension
        return versioning_extension.versionables_by_content[self.toolbar.obj.__class__]

    def _is_versioned(self):
        """Helper method to check if the model has been registered for
        versioning
        """
        versioning_extension = apps.get_app_config("djangocms_versioning").cms_extension
        return versioning_extension.is_content_model_versioned(
            self.toolbar.obj.__class__
        )

    def _get_proxy_model(self):
        """Helper method to get the proxy model class for the content
        model class
        """
        return self._get_versionable().version_model_proxy

    def _add_publish_button(self):
        """Helper method to add a publish button to the toolbar
        """
        # Check if object is registered with versioning otherwise don't add
        if not self._is_versioned():
            return
        # Add the publish button if in edit mode
        if self.toolbar.edit_mode_active:
            item = ButtonList(side=self.toolbar.RIGHT)
            proxy_model = self._get_proxy_model()
            version = Version.objects.get_for_content(self.toolbar.obj)
            publish_url = reverse(
                "admin:{app}_{model}_publish".format(
                    app=proxy_model._meta.app_label, model=proxy_model.__name__.lower()
                ),
                args=(version.pk,),
            )
            item.add_button(
                _("Publish"),
                url=publish_url,
                disabled=False,
                extra_classes=["cms-btn-action", "js-action", "cms-form-post-method", "cms-versioning-js-publish-btn"],
            )
            self.toolbar.add_item(item)

    def add_edit_button(self):
        """
        Only override the CMS versioning button when the object is versionable
        """
        if not self._is_versioned():
            # Show the standard cms edit button for non versionable objects
            return super().add_edit_button()
        self._add_edit_button()
        self._add_unlock_button()

    def _add_edit_button(self, disabled=False):
        """Helper method to add an edit button to the toolbar
        """
        item = ButtonList(side=self.toolbar.RIGHT)
        proxy_model = self._get_proxy_model()
        version = Version.objects.get_for_content(self.toolbar.obj)
        if version.check_edit_redirect.as_bool(self.request.user):
            edit_url = reverse(
                "admin:{app}_{model}_edit_redirect".format(
                    app=proxy_model._meta.app_label, model=proxy_model.__name__.lower()
                ),
                args=(version.pk,),
            )
            pks_for_grouper = version.versionable.for_content_grouping_values(
                version.content
            ).values_list("pk", flat=True)
            content_type = ContentType.objects.get_for_model(version.content)
            draft_exists = Version.objects.filter(
                object_id__in=pks_for_grouper, content_type=content_type, state=DRAFT
            ).exists()
            item.add_button(
                _("Edit") if draft_exists else _("New Draft"),
                url=edit_url,
                disabled=disabled,
                extra_classes=["cms-btn-action", "js-action", "cms-form-post-method", "cms-versioning-js-edit-btn"],
            )
            self.toolbar.add_item(item)

    def _add_unlock_button(self):
        """Helper method to add an edit button to the toolbar
        """
        if LOCK_VERSIONS and self._is_versioned():
            item = ButtonList(side=self.toolbar.RIGHT)
            proxy_model = self._get_proxy_model()
            version = Version.objects.get_for_content(self.toolbar.obj)
            if version.check_unlock.as_bool(self.request.user):
                unlock_url = reverse(
                    "admin:{app}_{model}_unlock".format(
                        app=proxy_model._meta.app_label, model=proxy_model.__name__.lower()
                    ),
                    args=(version.pk,),
                )
                can_unlock = self.request.user.has_perm("djangocms_versioning.delete_versionlock")
                if can_unlock:
                    extra_classes = [
                        "cms-btn-action",
                        "js-action",
                        "cms-form-post-method",
                        "cms-versioning-js-unlock-btn",
                    ]
                else:
                    extra_classes = ["cms-versioning-js-unlock-btn"]
                item.add_button(
                    _("Unlock"),
                    url=unlock_url if can_unlock else "#",
                    disabled=not can_unlock,
                    extra_classes=extra_classes,
                )
                self.toolbar.add_item(item)

    def _add_lock_message(self):
        if self._is_versioned() and LOCK_VERSIONS and not self.toolbar.edit_mode_active:
            version = Version.objects.get_for_content(self.toolbar.obj)
            lock_message = TemplateItem(
                template="djangocms_versioning/admin/lock_indicator.html",
                extra_context={"version": version},
                side=RIGHT,
            )
            self.toolbar.add_item(lock_message, position=0)

    def _add_revert_button(self, disabled=False):
        """Helper method to add a revert button to the toolbar
         """
        # Check if object is registered with versioning otherwise don't add
        if not self._is_versioned():
            return
        item = ButtonList(side=self.toolbar.RIGHT)
        proxy_model = self._get_proxy_model()
        version = Version.objects.get_for_content(self.toolbar.obj)
        if version.check_revert.as_bool(self.request.user):
            revert_url = reverse(
                "admin:{app}_{model}_revert".format(
                    app=proxy_model._meta.app_label,
                    model=proxy_model._meta.model_name,
                ),
                args=(version.pk,),
            )
            item.add_button(
                _("Revert"),
                url=revert_url,
                disabled=disabled,
                extra_classes=["cms-btn-action"],
            )
            self.toolbar.add_item(item)

    def _add_versioning_menu(self):
        """ Helper method to add version menu in the toolbar
        """
        # Check if object is registered with versioning otherwise don't add
        if not self._is_versioned():
            return

        version = Version.objects.get_for_content(self.toolbar.obj)
        if version is None:
            return

        version_menu_label = version.short_name()
        versioning_menu = self.toolbar.get_or_create_menu(
            VERSIONING_MENU_IDENTIFIER, version_menu_label, disabled=False
        )
        version = version.convert_to_proxy()
        if self.request.user.has_perm(
            "{app_label}.{codename}".format(
                app_label=version._meta.app_label,
                codename=get_permission_codename("change", version._meta),
            )
        ):
            url = version_list_url(version.content)
            versioning_menu.add_sideframe_item(_("Manage Versions"), url=url)
            # Compare to source menu entry
            if version.source:
                name = _("Compare to {source}").format(source=_(version.source.short_name()))
                proxy_model = self._get_proxy_model()
                url = reverse("admin:{app}_{model}_compare".format(
                     app=proxy_model._meta.app_label, model=proxy_model.__name__.lower()
                ), args=(version.source.pk,))

                url += "?" + urlencode({
                    "compare_to": version.pk,
                    "back": self.request.get_full_path(),
                })
                versioning_menu.add_link_item(name, url=url)
                # Discard changes menu entry (wrt to source)
                if version.check_discard.as_bool(self.request.user):  # pragma: no cover
                    versioning_menu.add_item(Break())
                    versioning_menu.add_link_item(
                        _("Discard Changes"),
                        url=reverse("admin:{app}_{model}_discard".format(
                            app=proxy_model._meta.app_label, model=proxy_model.__name__.lower()
                        ), args=(version.pk,))
                    )

    def _get_published_page_version(self):
        """Returns a published page if one exists for the toolbar object
        """
        language = self.current_lang

        # Exit the current toolbar object is not a Page / PageContent instance
        if not isinstance(self.toolbar.obj, PageContent) or not self.page:
            return

        return PageContent._original_manager.filter(
            page=self.page, language=language, versions__state=PUBLISHED
        ).first()

    def _add_view_published_button(self):
        """Helper method to add a publish button to the toolbar
        """
        # Check if object is registered with versioning otherwise don't add
        if not self._is_versioned():
            return

        # Add the View published button if in edit or preview mode
        published_version = self._get_published_page_version()
        if not published_version:
            return

        url = published_version.get_absolute_url() if hasattr(published_version, "get_absolute_url") else None
        if url and (self.toolbar.edit_mode_active or self.toolbar.preview_mode_active):
            item = ButtonList(side=self.toolbar.RIGHT)
            item.add_button(
                _("View Published"),
                url=url,
                disabled=False,
                extra_classes=["cms-btn", "cms-btn-switch-save"],
            )
            self.toolbar.add_item(item)

    def _add_preview_button(self):
        """Helper method to add a preview button to the toolbar when not in preview mode"""
        # Check if object is registered with versioning otherwise don't add
        if not self._is_versioned():
            return

        if not self.toolbar.preview_mode_active and not self.toolbar.edit_mode_active:
            # Any mode not preview mode can have a preview button
            # Exclude edit mode, however, since the django CMS core already ads the preview button for edit mode
            self.add_preview_button()

    def post_template_populate(self):
        super().post_template_populate()
        self._add_lock_message()
        self._add_preview_button()
        self._add_view_published_button()
        self._add_revert_button()
        self._add_publish_button()
        self._add_versioning_menu()


class VersioningPageToolbar(PageToolbar):
    """
    Overriding the original Page toolbar to ensure that draft and published pages
    can be accessed and to allow full control over the Page toolbar for versioned pages.
    """
    def get_page_content(self, language=None):
        if not language:
            language = self.current_lang

        return get_latest_admin_viewable_content(self.page, language=language)

    def populate(self):
        self.page = self.request.current_page or getattr(self.toolbar.obj, "page", None)
        self.title = self.get_page_content() if self.page else None
        self.permissions_activated = get_cms_setting("PERMISSION")

        self.override_language_menu()
        self.change_admin_menu()
        self.add_page_menu()
        self.change_language_menu()

    def override_language_menu(self):
        """
        Override the default language menu for pages that are versioned.
        The default language menu is too generic so for pages we need to replace it.
        """
        # Only override the menu if it exists and a page can be found
        language_menu = self.toolbar.get_menu(LANGUAGE_MENU_IDENTIFIER, _("Language"))
        if settings.USE_I18N and language_menu and self.page:
            # remove_item uses `items` attribute so we have to copy object
            for _item in copy(language_menu.items):
                language_menu.remove_item(item=_item)

            for code, name in get_language_tuple(self.current_site.pk):
                # Get the page content, it could be draft too!
                page_content = self.get_page_content(language=code)
                if page_content:
                    url = get_object_preview_url(page_content, code)
                    language_menu.add_link_item(name, url=url, active=self.current_lang == code)

    def change_language_menu(self):
        if self.toolbar.edit_mode_active and self.page:
            can_change = page_permissions.user_can_change_page(
                user=self.request.user, page=self.page, site=self.current_site
            )
        else:
            can_change = False

        if can_change:
            language_menu = self.toolbar.get_menu(LANGUAGE_MENU_IDENTIFIER)
            if not language_menu:
                return None

            languages = get_language_dict(self.current_site.pk)
            remove = [
                (code, languages.get(code, code))
                for code in self.page.get_languages()
                if code in languages
            ]
            add = [
                code
                for code in languages.items()
                if code not in remove
            ]
            copy = [
                (code, name)
                for code, name in languages.items()
                if code != self.current_lang and (code, name) in remove
            ]

            if add:
                language_menu.add_break(ADD_PAGE_LANGUAGE_BREAK)

                add_plugins_menu = language_menu.get_or_create_menu(
                    f"{LANGUAGE_MENU_IDENTIFIER}-add", _("Add Translation")
                )

                page_add_url = admin_reverse("cms_pagecontent_add")

                for code, name in add:
                    url = add_url_parameters(
                        page_add_url, cms_page=self.page.pk, language=code
                    )
                    add_plugins_menu.add_modal_item(name, url=url)

            if copy:
                copy_plugins_menu = language_menu.get_or_create_menu(
                    f"{LANGUAGE_MENU_IDENTIFIER}-copy", _("Copy all plugins")
                )
                title = _("from %s")
                question = _("Are you sure you want to copy all plugins from %s?")
                item_added = False
                for code, name in copy:
                    # Get the Draft or Published PageContent.
                    page_content = self.get_page_content(language=code)
                    if page_content:  # Only offer to copy if content for source language exists
                        page_copy_url = admin_reverse("cms_pagecontent_copy_language", args=(page_content.pk,))
                        copy_plugins_menu.add_ajax_item(
                            title % name, action=page_copy_url,
                            data={"source_language": code, "target_language": self.current_lang},
                            question=question % name, on_success=self.toolbar.REFRESH_PAGE
                        )
                        item_added = True
                    if not item_added:  # pragma: no cover
                        copy_plugins_menu.add_link_item(
                            _("No other language available"),
                            url="#",
                            disabled=True,
                        )


def replace_toolbar(old, new):
    """Replace `old` toolbar class with `new` class,
    while keeping its position in toolbar_pool.
    """
    new_name = ".".join((new.__module__, new.__name__))
    old_name = ".".join((old.__module__, old.__name__))
    toolbar_pool.toolbars = OrderedDict(
        (new_name, new) if name == old_name else (name, toolbar)
        for name, toolbar in toolbar_pool.toolbars.items()
    )


replace_toolbar(PageToolbar, VersioningPageToolbar)
replace_toolbar(PlaceholderToolbar, VersioningToolbar)



from cms.extensions.extension_pool import ExtensionPool
from cms.test_utils.testcases import CMSTestCase
from cms.utils.urlutils import admin_reverse
from django.contrib import admin
from django.contrib.sites.models import Site
from django.test import RequestFactory

from djangocms_versioning.cms_config import copy_page_content
from djangocms_versioning.models import Version
from djangocms_versioning.test_utils.extended_polls.admin import (
    PollExtensionAdmin,
)
from djangocms_versioning.test_utils.extended_polls.models import (
    PollPageContentExtension,
)
from djangocms_versioning.test_utils.extensions.models import (
    TestPageContentExtension,
    TestPageExtension,
)
from djangocms_versioning.test_utils.factories import (
    PageContentFactory,
    PageVersionFactory,
    PollTitleExtensionFactory,
    TestTitleExtensionFactory,
)


class ExtensionTestCase(CMSTestCase):
    def setUp(self):
        self.version = PageVersionFactory(content__language="en")
        de_pagecontent = PageContentFactory(
            page=self.version.content.page, language="de"
        )
        self.page = self.version.content.page
        site = Site.objects.first()
        self.new_page = self.page.copy(
            site=site,
            parent_node=self.page.node.parent,
            translations=False,
            permissions=False,
            extensions=False,
            user=self.get_superuser(),
        )
        new_page_content = PageContentFactory(page=self.new_page, language="de")
        self.new_page.page_content_cache[de_pagecontent.language] = new_page_content

    def test_copy_extensions(self):
        """Try to copy the extension, without the monkeypatch this tests fails"""
        extension_pool = ExtensionPool()
        extension_pool.page_extensions = {TestPageExtension}
        extension_pool.title_extensions = {TestPageContentExtension}
        extension_pool.copy_extensions(
            self.page, self.new_page, languages=["de"]
        )
        # No asserts, this test originally failed because the versioned manager was called
        # in copy_extensions, now we call the original manager instead
        # https://github.com/divio/djangocms-versioning/pull/201/files#diff-fc33dd7b5aa9b1645545cf48dfc9b4ecR19

    def test_pagecontent_copy_method_creates_extension_title_extension_attached(self):
        """
        The page content copy method should create a new title extension, if one is attached to it.
        """
        page_content = self.version.content
        poll_extension = PollTitleExtensionFactory(extended_object=page_content)
        poll_extension.votes = 5

        new_pagecontent = copy_page_content(page_content)

        self.assertNotEqual(new_pagecontent.pollpagecontentextension, poll_extension)
        self.assertEqual(page_content.pollpagecontentextension.pk, poll_extension.pk)
        self.assertNotEqual(page_content.pollpagecontentextension.pk, new_pagecontent.pollpagecontentextension.pk)
        self.assertEqual(new_pagecontent.pollpagecontentextension.votes, 5)
        self.assertEqual(PollPageContentExtension._base_manager.count(), 2)

    def test_pagecontent_copy_method_not_created_extension_title_extension_attached(self):
        """
        The pagecontent copy method should not create a new title extension, if one isn't attached to the pagecontent
        being copied
        """
        new_pagecontent = copy_page_content(self.version.content)

        self.assertFalse(hasattr(new_pagecontent, "polltitleextension"))
        self.assertEqual(PollPageContentExtension._base_manager.count(), 0)

    def test_pagecontent_copy_method_creates_extension_multiple_title_extension_attached(self):
        """
        The page content copy method should handle creation of multiple extensions
        """
        page_content = self.version.content
        poll_extension = PollTitleExtensionFactory(extended_object=page_content)
        poll_extension.votes = 5
        title_extension = TestTitleExtensionFactory(extended_object=page_content)

        new_pagecontent = copy_page_content(page_content)

        self.assertNotEqual(new_pagecontent.pollpagecontentextension, poll_extension)
        self.assertEqual(page_content.pollpagecontentextension.pk, poll_extension.pk)
        self.assertNotEqual(new_pagecontent.testpagecontentextension, poll_extension)
        self.assertEqual(page_content.testpagecontentextension.pk, poll_extension.pk)
        self.assertNotEqual(new_pagecontent.pollpagecontentextension, poll_extension)
        self.assertNotEqual(new_pagecontent.testpagecontentextension, title_extension)
        self.assertEqual(new_pagecontent.pollpagecontentextension.votes, 5)
        self.assertEqual(PollPageContentExtension._base_manager.count(), 2)
        self.assertEqual(TestPageContentExtension._base_manager.count(), 2)

    def test_title_extension_admin_monkey_patch_save(self):
        """
        When hitting the monkeypatched save method, with a draft pagecontent, ensure that we don't see failures
        due to versioning overriding monkeypatches
        """
        poll_extension = PollTitleExtensionFactory(extended_object=self.version.content)
        model_site = PollExtensionAdmin(admin_site=admin.AdminSite(), model=PollPageContentExtension)
        test_url = admin_reverse("extended_polls_pollpagecontentextension_change", args=(poll_extension.pk,))
        test_url += "?extended_object=%s" % self.version.content.pk
        request = RequestFactory().post(path=test_url)
        request.user = self.get_superuser()

        poll_extension.votes = 1
        model_site.save_model(request, poll_extension, form=None, change=False)

        self.assertEqual(PollPageContentExtension._base_manager.first().votes, 1)
        self.assertEqual(PollPageContentExtension._base_manager.count(), 1)

    def test_title_extension_admin_monkey_patch_save_date_modified_updated(self):
        """
        When making changes to an extended model that is attached to a PageContent via
        the Title Extension the date modified in a version should also be updated to reflect
        the correct date timestamp.
        """
        poll_extension = PollTitleExtensionFactory(extended_object=self.version.content)
        model_site = PollExtensionAdmin(admin_site=admin.AdminSite(), model=PollPageContentExtension)
        pre_changes_date_modified = Version.objects.get(id=self.version.pk).modified
        test_url = admin_reverse("extended_polls_pollpagecontentextension_change", args=(poll_extension.pk,))
        test_url += "?extended_object=%s" % self.version.content.pk

        request = RequestFactory().post(path=test_url)
        request.user = self.get_superuser()
        model_site.save_model(request, poll_extension, form=None, change=False)

        post_changes_date_modified = Version.objects.get(id=self.version.pk).modified

        self.assertNotEqual(pre_changes_date_modified, post_changes_date_modified)

    def test_title_extension_admin_monkeypatch_add_view(self):
        """
        When hitting the add view, without the monkeypatch, the pagecontent queryset will be filtered to only show
        published. Hit it with a draft, to make sure the monkeypatch works.
        """
        with self.login_user_context(self.get_superuser()):
            response = self.client.get(
                admin_reverse("extended_polls_pollpagecontentextension_add") +
                "?extended_object=%s" % self.version.content.pk,
                follow=True
            )
            self.assertEqual(response.status_code, 200)



from cms.test_utils.testcases import CMSTestCase

from djangocms_versioning.constants import DRAFT
from djangocms_versioning.models import Version
from djangocms_versioning.test_utils.polls.models import Poll, PollContent


class CheckDraftEditableTestCase(CMSTestCase):
    def test_creation_wo_with_user(self):

        poll = Poll.objects.create(name="my test poll")
        poll_content = PollContent.objects.create(poll=poll, language="en", text="Do you love django CMS?")
        version = Version.objects.create(content=poll_content, created_by=self.get_superuser())
        self.assertEqual(version.state, DRAFT)
        self.assertTrue(poll_content.versions.exists())

    def test_creation_with_user(self):
        poll = Poll.objects.create(name="my test poll")
        user = self.get_superuser()
        poll_content = PollContent.objects\
            .with_user(user=user).create(poll=poll, language="en", text="Do you love django CMS?")
        self.assertTrue(poll_content.versions.exists())



from cms.test_utils.testcases import CMSTestCase
from django.core.management import call_command
from django.db import transaction

from djangocms_versioning import constants
from djangocms_versioning.models import Version
from djangocms_versioning.test_utils.blogpost.models import (
    BlogContent,
    BlogPost,
)
from djangocms_versioning.test_utils.polls.models import Poll, PollContent


class CreateVersionsTestCase(CMSTestCase):
    def test_create_versions(self):
        content_models_by_language = {"en": 5, "de": 2, "nl": 7}

        # Arrange:
        # Create BlogPosts and Poll w/o versioned content objects
        with transaction.atomic():
            post = BlogPost(name="my multi-lingual blog post")
            post.save()
            for language, cnt in content_models_by_language.items():
                for _i in range(cnt):
                    # Use save NOT objects.create to avoid creating Version object
                    BlogContent(blogpost=post, language=language).save()
            poll = Poll()
            poll.save()
            for language, cnt in content_models_by_language.items():
                for _i in range(cnt):
                    # Use save NOT objects.create to avoid creating Version object
                    PollContent(poll=poll, language=language).save()
        # Verify that no Version objects have been created
        self.assertEqual(Version.objects.count(), 0)

        # Act:
        # Call create_versions command
        try:
            call_command("create_versions", userid=self.get_superuser().pk, state=constants.DRAFT)
        except SystemExit as e:
            status_code = str(e)
        else:
            # the "no changes" exit code is 0
            status_code = "0"
        self.assertEqual(status_code, "0")

        # Assert:
        # Blog has no additional grouping field, i.e. all except the last blog content must be archived
        blog_contents = BlogContent.admin_manager.filter(blogpost=post, language=language).order_by("-pk")
        self.assertEqual(blog_contents[0].versions.first().state, constants.DRAFT)
        for cont in blog_contents[1:]:
            self.assertEqual(cont.versions.first().state, constants.ARCHIVED)

        # Poll has additional grouping field, i.e. for each language there must be one draft (rest archived)
        for language, _cnt in content_models_by_language.items():
            poll_contents = PollContent.admin_manager.filter(poll=poll, language=language).order_by("-pk")
            self.assertEqual(poll_contents[0].versions.first().state, constants.DRAFT)
            for cont in poll_contents[1:]:
                self.assertEqual(cont.versions.first().state, constants.ARCHIVED)



# Generated by Django 3.2.16 on 2022-10-28 22:33

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('cms', '0034_remove_pagecontent_placeholders'),
    ]

    operations = [
        migrations.CreateModel(
            name='Poll',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.TextField()),
            ],
        ),
        migrations.CreateModel(
            name='PollPlugin',
            fields=[
                ('cmsplugin_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, related_name='polls_pollplugin', serialize=False, to='cms.cmsplugin')),
                ('poll', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='polls.poll')),
            ],
            bases=('cms.cmsplugin',),
        ),
        migrations.CreateModel(
            name='PollContent',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('language', models.TextField()),
                ('text', models.TextField()),
                ('poll', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='polls.poll')),
            ],
        ),
        migrations.CreateModel(
            name='Answer',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('text', models.TextField()),
                ('poll_content', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='polls.pollcontent')),
            ],
        ),
    ]



from collections import OrderedDict
from unittest.mock import Mock, patch

from cms.admin.forms import ChangePageForm
from cms.models import Page
from cms.test_utils.testcases import CMSTestCase
from django.apps import apps
from django.contrib import admin
from django.contrib.sites.models import Site
from django.core.exceptions import ImproperlyConfigured
from django.test import RequestFactory
from django.utils.text import slugify

from djangocms_versioning.admin import VersionAdmin, VersioningAdminMixin
from djangocms_versioning.cms_config import (
    VersioningCMSConfig,
    VersioningCMSExtension,
)
from djangocms_versioning.constants import DRAFT
from djangocms_versioning.datastructures import VersionableItem, default_copy
from djangocms_versioning.models import Version
from djangocms_versioning.test_utils import factories
from djangocms_versioning.test_utils.blogpost.cms_config import (
    BlogpostCMSConfig,
)
from djangocms_versioning.test_utils.blogpost.models import (
    BlogContent,
    Comment,
)
from djangocms_versioning.test_utils.incorrectly_configured_blogpost.cms_config import (
    IncorrectBlogpostCMSConfig,
)
from djangocms_versioning.test_utils.polls.cms_config import PollsCMSConfig
from djangocms_versioning.test_utils.polls.models import Poll, PollContent

req_factory = RequestFactory()


class PageContentVersioningBehaviourTestCase(CMSTestCase):

    def setUp(self):
        self.site = Site.objects.get_current()
        self.user = self.get_superuser()
        self.language = "en"
        self.title = "test page"

        self.version = factories.PageVersionFactory(content__language="en", state=DRAFT,)
        factories.PageUrlFactory(
            page=self.version.content.page,
            language="en",
            path=slugify(self.title),
            slug=slugify(self.title),
        )

        self.page = self.version.content.page
        self.content = self.version.content

    def test_saving_draft_has_pageurl(self):
        """A draft has pageurl as normal"""
        url = self.page.get_urls().first()

        self.assertEqual(self.version.state, DRAFT)
        self.assertEqual(url.path, slugify(self.title))

    def test_publishing_draft_retains_pageurl(self):
        """We publish a version, pageurl is still the same"""
        self.version.publish(self.user)

        url = self.page.get_urls().first()
        self.assertEqual(url.path, slugify(self.title))

    def test_published_version_with_new_version_retains_pageurl(self):
        """
        We have a published page and we create a new version and publish that.
        PageUrl path stays the same
        """
        self.version.publish(self.user)
        v2 = self.version.copy(self.user)
        v2.publish(self.user)

        page = Page.objects.get(pk=self.page.pk)
        url = page.get_urls().first()

        self.assertEqual(url.path, slugify(self.title))

    def test_published_version_with_new_version_retains_pageurl_unmanaged(self):
        """We have a published page and we create a new version and publish that.
        The url in question is unmanaged and stays the same.

        The path was previously set to None on unpublish. This deleted the
        url a content editor had put in the field "overwrite url".
        Putting anything into the field "overwrite url" sets the property
        managed to False.
        """
        url = self.page.get_urls().first()
        url.managed = False
        url.save()

        self.version.publish(self.user)
        v2 = self.version.copy(self.user)
        v2.publish(self.user)

        page = Page.objects.get(pk=self.page.pk)
        url = page.get_urls().first()

        self.assertEqual(url.path, slugify(self.title))

    def test_changing_slug_changes_page_url(self):
        """Using change form to change title / slug updates path?"""
        new_title = "new slug here"
        data = {
            "title": self.content.title,
            "slug": new_title
        }

        request = req_factory.get("/?language=en")
        request.user = self.user

        form = ChangePageForm(data, instance=self.content)
        form._request = request
        form._site = self.site
        self.assertEqual(form.is_valid(), True)

        form.save()
        page = Page.objects.get(pk=self.page.pk)
        url = page.get_urls().first()

        self.assertEqual(url.slug, slugify(new_title))
        self.assertEqual(url.path, slugify(new_title))


class VersioningExtensionUnitTestCase(CMSTestCase):
    def test_raises_exception_if_neither_versioning_nor_context_setting_specified(self):
        """
        Tests, if neither the versioning attribute or the
        versioning_add_to_confirmation_context attribute has been specified,
        an ImproperlyConfigured exception is raised
        """
        extensions = VersioningCMSExtension()
        cms_config = Mock(spec=[], djangocms_versioning_enabled=True)
        with self.assertRaises(ImproperlyConfigured):
            extensions.configure_app(cms_config)

    def test_raises_exception_if_versioning_not_iterable(self):
        """Tests ImproperlyConfigured exception is raised if
        versioning setting is not an iterable
        """
        extensions = VersioningCMSExtension()
        cms_config = Mock(
            spec=[],
            djangocms_versioning_enabled=True,
            versioning=VersionableItem(
                content_model=PollContent,
                grouper_field_name="poll",
                copy_function=default_copy,
            ),
        )
        with self.assertRaises(ImproperlyConfigured):
            extensions.configure_app(cms_config)

    def test_raises_exception_if_not_versionable_class(self):
        """Tests ImproperlyConfigured exception is raised if elements
        in the versioning list are not instances of VersionableItem classes
        """
        extensions = VersioningCMSExtension()
        cms_config = Mock(
            spec=[], djangocms_versioning_enabled=True, versioning=["aaa", {}]
        )
        with self.assertRaises(ImproperlyConfigured):
            extensions.configure_app(cms_config)

    def test_raises_exception_if_content_class_already_registered_in_same_config(self):
        """Tests ImproperlyConfigured exception is raised if the same
        content class is registered twice in the same config file
        """
        extension = VersioningCMSExtension()
        poll_versionable = VersionableItem(
            content_model=PollContent,
            grouper_field_name="poll",
            copy_function=default_copy,
        )
        poll_versionable2 = VersionableItem(
            content_model=PollContent,
            grouper_field_name="poll",
            copy_function=default_copy,
        )
        cms_config = Mock(
            spec=[],
            djangocms_versioning_enabled=True,
            versioning=[poll_versionable, poll_versionable2],
        )
        with self.assertRaises(ImproperlyConfigured):
            extension.configure_app(cms_config)

    def test_raises_exception_if_content_class_already_registered_in_different_config(
        self
    ):
        """Tests ImproperlyConfigured exception is raised if the same
        content class is registered twice in different config files
        """
        extension = VersioningCMSExtension()
        poll_versionable = VersionableItem(
            content_model=PollContent,
            grouper_field_name="poll",
            copy_function=default_copy,
        )
        poll_versionable2 = VersionableItem(
            content_model=PollContent,
            grouper_field_name="poll",
            copy_function=default_copy,
        )
        cms_config1 = Mock(
            spec=[], djangocms_versioning_enabled=True, versioning=[poll_versionable]
        )
        cms_config2 = Mock(
            spec=[], djangocms_versioning_enabled=True, versioning=[poll_versionable2]
        )
        with self.assertRaises(ImproperlyConfigured):
            extension.handle_versioning_setting(cms_config1)
            extension.handle_versioning_setting(cms_config2)

    def test_versioning_add_to_confirmation_context_is_an_optional_setting(self):
        extension = VersioningCMSExtension()
        poll_versionable = VersionableItem(
            content_model=PollContent,
            grouper_field_name="poll",
            copy_function=default_copy,
        )
        cms_config = Mock(
            spec=[], djangocms_versioning_enabled=True, versioning=[poll_versionable]
        )
        try:
            extension.configure_app(cms_config)
        except ImproperlyConfigured:
            self.fail(
                "versioning_add_to_confirmation_context setting should be optional"
            )

    def test_raises_exception_if_unsupported_key_added_to_add_to_context(self):
        """Tests ImproperlyConfigured exception is raised if an unsupported
        dict key is used for the versioning_add_to_confirmation_context setting
        """
        extension = VersioningCMSExtension()
        poll_versionable = VersionableItem(
            content_model=PollContent,
            grouper_field_name="poll",
            copy_function=default_copy,
        )
        cms_config = Mock(
            spec=[],
            djangocms_versioning_enabled=True,
            versioning=[poll_versionable],
            # versioning doesn't know what red rabbits is
            # so this should raise an exception
            versioning_add_to_confirmation_context={
                "red_rabbits": OrderedDict({"rabbit": lambda r, v: v.content})
            },
        )
        with self.assertRaises(ImproperlyConfigured):
            extension.configure_app(cms_config)

    def test_versionables_list_created(self):
        """Test handle_versioning_setting method adds all the
        models into the versionables list
        """
        extension = VersioningCMSExtension()
        poll_versionable = VersionableItem(
            content_model=PollContent,
            grouper_field_name="poll",
            copy_function=default_copy,
        )
        blog_versionable = VersionableItem(
            content_model=BlogContent,
            grouper_field_name="blogpost",
            copy_function=default_copy,
        )
        cms_config = Mock(
            spec=[],
            djangocms_versioning_enabled=True,
            versioning=[poll_versionable, blog_versionable],
        )
        extension.configure_app(cms_config)
        self.assertListEqual(
            extension.versionables, [poll_versionable, blog_versionable]
        )

    def test_context_dict_created(self):
        """Test a dict is populated from the versioning_add_to_confirmation_context
        config
        """
        extension = VersioningCMSExtension()

        def unpublish_context1(request, version, *args, **kwargs):
            return "Every time you unpublish something you should kiss a cat"

        def unpublish_context2(request, version, *args, **kwargs):
            return "Good luck with your unpublishing"

        cms_config = Mock(
            spec=[],
            djangocms_versioning_enabled=True,
            versioning_add_to_confirmation_context={
                "unpublish": OrderedDict(
                    {"1": unpublish_context1, "2": unpublish_context2}
                )
            },
        )
        extension.configure_app(cms_config)
        expected = {
            "unpublish": OrderedDict({"1": unpublish_context1, "2": unpublish_context2})
        }
        self.assertDictEqual(extension.add_to_context, expected)

    def test_context_dict_doesnt_get_overwritten(self):
        """Test when multiple apps update the same key in the context dict,
        values from both apps end up under that key
        """
        extension = VersioningCMSExtension()

        def unpublish_context1(request, version, *args, **kwargs):
            return "Cats don't like to be unpublished"

        def unpublish_context2(request, version, *args, **kwargs):
            return "Elephants don't mind being unpublished'"

        cms_config1 = Mock(
            spec=[],
            djangocms_versioning_enabled=True,
            versioning_add_to_confirmation_context={
                "unpublish": OrderedDict([("1", unpublish_context1)])
            },
        )
        cms_config2 = Mock(
            spec=[],
            djangocms_versioning_enabled=True,
            versioning_add_to_confirmation_context={
                "unpublish": OrderedDict([("2", unpublish_context2)])
            },
        )

        extension.configure_app(cms_config1)
        extension.configure_app(cms_config2)

        expected = {
            "unpublish": OrderedDict(
                [("1", unpublish_context1), ("2", unpublish_context2)]
            )
        }
        self.assertDictEqual(extension.add_to_context, expected)

    def test_handle_content_admin_classes(self):
        """Test handle_admin_classes replaces the admin model class
        with an admin model class that inherits from VersioningAdminMixin
        """
        extensions = VersioningCMSExtension()
        cms_config = Mock(
            spec=[],
            djangocms_versioning_enabled=True,
            versioning=[
                VersionableItem(
                    content_model=PollContent,
                    grouper_field_name="poll",
                    copy_function=default_copy,
                )
            ],
        )
        extensions.handle_admin_classes(cms_config)
        self.assertIn(PollContent, admin.site._registry)
        self.assertIn(
            VersioningAdminMixin, admin.site._registry[PollContent].__class__.mro()
        )

    def test_is_content_model_versioned(self):
        """Test that is_content_model_versioned returns True for
        content model that's versioned
        """
        extension = VersioningCMSExtension()
        extension.versionables = [
            VersionableItem(
                content_model=PollContent,
                grouper_field_name="poll",
                copy_function=default_copy,
            )
        ]

        self.assertTrue(extension.is_content_model_versioned(PollContent))

    def test_is_content_model_not_versioned(self):
        """Test that is_content_model_versioned returns False for
        content model that's not versioned
        """
        extension = VersioningCMSExtension()
        extension.versionables = []

        self.assertFalse(extension.is_content_model_versioned(PollContent))

    def test_handle_version_admin(self):
        versionable = Mock(
            spec=[],
            djangocms_versioning_enabled=True,
            content_model=PollContent,
            concrete=True,
            grouper_model=Poll,
            version_model_proxy=apps.get_model(
                "djangocms_versioning", "PollContentVersion"
            ),
        )

        with patch.object(versionable, "version_model_proxy"):
            extensions = VersioningCMSExtension()
            cms_config = Mock(
                spec=[], djangocms_versioning_enabled=True, versioning=[versionable]
            )
            extensions.handle_version_admin(cms_config)
        self.assertIn(versionable.version_model_proxy, admin.site._registry)
        self.assertIn(
            VersionAdmin,
            admin.site._registry[versionable.version_model_proxy].__class__.mro(),
        )

    def test_field_extension_populates(self):
        """
        With proper configuration provided, cms extension populates
        """
        def poll_modifier(obj, field):
            return obj

        extensions = VersioningCMSExtension()
        cms_config = Mock(
            spec=[],
            djangocms_versioning_enabled=True,
            versioning=[
                VersionableItem(
                    content_model=PollContent,
                    grouper_field_name="poll",
                    copy_function=default_copy,
                )
            ],
            extended_admin_field_modifiers=[{PollContent: {"text": poll_modifier}}, ]
        )
        extensions.handle_admin_field_modifiers(cms_config)

        self.assertEqual(extensions.add_to_field_extension, {PollContent: {"text": poll_modifier}})

    def test_field_extension_proper_error_non_iterable(self):
        """
        When a non-iterable is passed as the method for modifying a field,
        raise ImproperlyConfigured
        """
        def poll_modifier(obj, field):
            return obj

        extensions = VersioningCMSExtension()
        cms_config = Mock(
            spec=[],
            djangocms_versioning_enabled=True,
            versioning=[
                VersionableItem(
                    content_model=PollContent,
                    grouper_field_name="poll",
                    copy_function=default_copy,
                )
            ],
            extended_admin_field_modifiers=(PollContent, "text", poll_modifier)
        )

        with self.assertRaises(ImproperlyConfigured):
            extensions.handle_admin_field_modifiers(cms_config)


# NOTE: These tests simply test what has already happened on start up
# when the app registry has been instantiated.
class VersioningIntegrationTestCase(CMSTestCase):
    def test_all_versionables_collected(self):
        """Check that all version models defined in cms_config.py
        are collected into a list
        """
        app = apps.get_app_config("djangocms_versioning")
        page_versionable = VersioningCMSConfig.versioning[0]
        poll_versionable = PollsCMSConfig.versioning[0]
        blog_versionable = BlogpostCMSConfig.versioning[0]
        comment_versionable = BlogpostCMSConfig.versioning[1]
        incorrect_blog_versionable = IncorrectBlogpostCMSConfig.versioning[0]
        self.assertListEqual(
            app.cms_extension.versionables,
            [page_versionable, poll_versionable, blog_versionable, comment_versionable, incorrect_blog_versionable],
        )

    def test_admin_classes_reregistered(self):
        """Integration test that all content models that are registered
        with the admin have their admin class overridden with a
        subclass of VersioningAdminMixin
        """
        # TODO: Awaiting FIL-313 to land on core
        # Check PageContent has had its admin class modified
        # self.assertIn(PageContent, admin.site._registry)
        # self.assertIn(
        #     VersioningAdminMixin,
        #     admin.site._registry[PageContent].__class__.mro()
        # )
        # Check PollContent has had its admin class modified
        self.assertIn(PollContent, admin.site._registry)
        self.assertIn(
            VersioningAdminMixin, admin.site._registry[PollContent].__class__.mro()
        )
        # Check BlogContent has had its admin class modified
        self.assertIn(BlogContent, admin.site._registry)
        self.assertIn(
            VersioningAdminMixin, admin.site._registry[BlogContent].__class__.mro()
        )
        # Check that Comments were not registered to the admin
        # (they are defined in cms_config.py but are not registered
        # to the admin)
        self.assertNotIn(Comment, admin.site._registry)

    def test_version_admin_registered(self):
        """Integration test that for each content model
        there's a version proxy registered with the admin
        subclassing VersionAdmin
        """
        version_proxies = [
            model for model in admin.site._registry if issubclass(model, Version)
        ]
        source_models_in_proxies = [model._source_model for model in version_proxies]
        source_model_to_proxy = dict(zip(source_models_in_proxies, version_proxies))

        self.assertIn(PollContent, source_models_in_proxies)
        self.assertIn(
            VersionAdmin,
            admin.site._registry[source_model_to_proxy[PollContent]].__class__.mro(),
        )

        self.assertIn(BlogContent, source_models_in_proxies)
        self.assertIn(
            VersionAdmin,
            admin.site._registry[source_model_to_proxy[BlogContent]].__class__.mro(),
        )

        self.assertNotIn(Comment, source_models_in_proxies)



# -*- coding: utf-8 -*-
# Generated by Django 1.11 on 2018-07-30 14:41
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ("contenttypes", "0002_remove_content_type_name"),
        ("djangocms_versioning", "0002_delete_campaign"),
    ]

    operations = [
        migrations.CreateModel(
            name="Version",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("label", models.TextField()),
                ("created", models.DateTimeField(auto_now_add=True, verbose_name="Created")),
                ("object_id", models.PositiveIntegerField()),
                (
                    "content_type",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        to="contenttypes.ContentType",
                    ),
                ),
            ],
        )
    ]



from django.contrib import admin

from djangocms_versioning.admin import ExtendedVersionAdminMixin

from .models import IncorrectBlogContent


class IncorrectBlogContentAdmin(ExtendedVersionAdminMixin, admin.ModelAdmin):
    pass


admin.site.register(IncorrectBlogContent, IncorrectBlogContentAdmin)



from cms.models import CMSPlugin
from django.db import models
from django.urls import reverse


class Poll(models.Model):
    name = models.TextField()

    def __str__(self):
        return f"{self.name} ({self.pk})"


class PollContent(models.Model):
    poll = models.ForeignKey(Poll, on_delete=models.CASCADE)
    language = models.TextField()
    text = models.TextField()

    def __str__(self):
        return self.text

    def get_absolute_url(self):
        return reverse("admin:polls_pollcontent_changelist")

    def get_preview_url(self):
        return reverse("admin:polls_pollcontent_preview", args=[self.id])


class Answer(models.Model):
    poll_content = models.ForeignKey(PollContent, on_delete=models.CASCADE)
    text = models.TextField()

    def __str__(self):
        return self.text


class PollPlugin(CMSPlugin):
    poll = models.ForeignKey(Poll, on_delete=models.CASCADE)

    def __str__(self):
        return str(self.poll)



# Generated by Django 3.2.19 on 2023-05-05 09:34

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('djangocms_versioning', '0015_version_modified'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='version',
            options={'permissions': (('delete_versionlock', 'Can unlock verision'),)},
        ),
        migrations.AddField(
            model_name='version',
            name='locked_by',
            field=models.ForeignKey(default=None, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='locking_users', to=settings.AUTH_USER_MODEL, verbose_name='locked by'),
        ),
    ]



import string

import factory
from cms import constants
from cms.models import Page, PageContent, PageUrl, Placeholder, TreeNode
from django.contrib.auth.models import User
from django.contrib.contenttypes.models import ContentType
from django.contrib.sites.models import Site
from djangocms_text_ckeditor.models import Text
from factory.fuzzy import FuzzyChoice, FuzzyInteger, FuzzyText

from ..models import Version
from .blogpost.models import BlogContent, BlogPost
from .extended_polls.models import PollPageContentExtension
from .extensions.models import TestPageContentExtension
from .incorrectly_configured_blogpost.models import (
    IncorrectBlogContent,
    IncorrectBlogPost,
)
from .polls.models import Answer, Poll, PollContent
from .unversioned_editable_app.models import FancyPoll


class UserFactory(factory.django.DjangoModelFactory):
    username = FuzzyText(length=12)
    first_name = factory.Faker("first_name")
    last_name = factory.Faker("last_name")
    email = factory.LazyAttribute(
        lambda u: f"{u.first_name.lower()}.{u.last_name.lower()}@example.com"
    )

    class Meta:
        model = User

    @classmethod
    def _create(cls, model_class, *args, **kwargs):
        """Override the default ``_create`` with our custom call."""
        manager = cls._get_manager(model_class)
        # The default would use ``manager.create(*args, **kwargs)``
        return manager.create_user(*args, **kwargs)


class AbstractVersionFactory(factory.django.DjangoModelFactory):
    object_id = factory.SelfAttribute("content.id")
    content_type = factory.LazyAttribute(
        lambda o: ContentType.objects.get_for_model(o.content)
    )
    created_by = factory.SubFactory(UserFactory)

    class Meta:
        exclude = ["content"]
        abstract = True


class AbstractContentFactory(factory.django.DjangoModelFactory):
    @classmethod
    def _get_manager(cls, model_class):
        return model_class._base_manager


class PollFactory(factory.django.DjangoModelFactory):
    name = FuzzyText(length=6)

    class Meta:
        model = Poll


class PollContentFactory(factory.django.DjangoModelFactory):
    poll = factory.SubFactory(PollFactory)
    language = FuzzyChoice(["en", "fr", "it"])
    text = FuzzyText(length=24)

    class Meta:
        model = PollContent

    @classmethod
    def _get_manager(cls, model_class):
        return model_class._base_manager


class PollVersionFactory(AbstractVersionFactory):
    content = factory.SubFactory(PollContentFactory)

    class Meta:
        model = Version


class PollContentWithVersionFactory(PollContentFactory):
    @factory.post_generation
    def version(self, create, extracted, **kwargs):
        # NOTE: Use this method as below to define version attributes:
        # PollContentWithVersionFactory(version__label='label1')
        if not create:
            # Simple build, do nothing.
            return
        PollVersionFactory(content=self, **kwargs)


class AnswerFactory(factory.django.DjangoModelFactory):
    poll_content = factory.SubFactory(PollContentFactory)
    text = factory.LazyAttributeSequence(
        lambda o, n: "Poll %s - Answer %d" % (o.poll_content.poll.name, n)
    )

    class Meta:
        model = Answer


class BlogPostFactory(factory.django.DjangoModelFactory):
    name = FuzzyText(length=6)

    class Meta:
        model = BlogPost


class BlogContentFactory(AbstractContentFactory):
    blogpost = factory.SubFactory(BlogPostFactory)
    language = FuzzyChoice(["en", "fr", "it"])
    text = FuzzyText(length=24)

    class Meta:
        model = BlogContent


class BlogPostVersionFactory(AbstractVersionFactory):
    content = factory.SubFactory(BlogContentFactory)

    class Meta:
        model = Version


class BlogContentWithVersionFactory(BlogContentFactory):
    @factory.post_generation
    def version(self, create, extracted, **kwargs):
        # NOTE: Use this method as below to define version attributes:
        # BlogContentWithVersionFactory(version__label='label1')
        if not create:
            # Simple build, do nothing.
            return
        BlogPostVersionFactory(content=self, **kwargs)


class IncorrectBlogPostFactory(factory.django.DjangoModelFactory):
    name = FuzzyText(length=6)

    class Meta:
        model = IncorrectBlogPost


class IncorrectBlogContentFactory(AbstractContentFactory):
    blogpost = factory.SubFactory(IncorrectBlogPostFactory)
    text = FuzzyText(length=24)

    class Meta:
        model = IncorrectBlogContent


class IncorrectBlogPostVersionFactory(AbstractVersionFactory):
    content = factory.SubFactory(IncorrectBlogContentFactory)

    class Meta:
        model = Version


class IncorrectBlogContentWithVersionFactory(IncorrectBlogContentFactory):
    @factory.post_generation
    def version(self, create, extracted, **kwargs):
        if not create:
            return
        IncorrectBlogPostVersionFactory(content=self, **kwargs)


class TreeNodeFactory(factory.django.DjangoModelFactory):
    site = factory.fuzzy.FuzzyChoice(Site.objects.all())
    depth = 0
    # NOTE: Generating path this way is probably not a good way of
    # doing it, but seems to work for our present tests which only
    # really need a tree node to exist and not throw unique constraint
    # errors on this field. If the data in this model starts mattering
    # in our tests then something more will need to be done here.
    path = FuzzyText(length=8, chars=string.digits)

    class Meta:
        model = TreeNode


class PageUrlFactory(factory.django.DjangoModelFactory):
    slug = ""
    path = ""
    managed = False
    language = "en"

    class Meta:
        model = PageUrl


class PageFactory(factory.django.DjangoModelFactory):
    node = factory.SubFactory(TreeNodeFactory)

    class Meta:
        model = Page


class PageContentFactory(AbstractContentFactory):
    page = factory.SubFactory(PageFactory)
    language = FuzzyChoice(["en", "fr", "it"])
    title = FuzzyText(length=12)
    page_title = FuzzyText(length=12)
    menu_title = FuzzyText(length=12)
    meta_description = FuzzyText(length=12)
    redirect = FuzzyText(length=12)
    created_by = FuzzyText(length=12)
    changed_by = FuzzyText(length=12)
    in_navigation = FuzzyChoice([True, False])
    soft_root = FuzzyChoice([True, False])
    limit_visibility_in_menu = constants.VISIBILITY_USERS
    template = "page.html"
    xframe_options = FuzzyInteger(0, 25)

    class Meta:
        model = PageContent


class PageVersionFactory(AbstractVersionFactory):
    content = factory.SubFactory(PageContentFactory)

    class Meta:
        model = Version


class PageContentWithVersionFactory(PageContentFactory):
    @factory.post_generation
    def version(self, create, extracted, **kwargs):
        # NOTE: Use this method as below to define version attributes:
        # PageContentWithVersionFactory(version__label='label1')
        if not create:
            # Simple build, do nothing.
            return
        PageVersionFactory(content=self, **kwargs)


class PlaceholderFactory(factory.django.DjangoModelFactory):
    default_width = FuzzyInteger(0, 25)
    slot = FuzzyText(length=2, chars=string.digits)
    # NOTE: When using this factory you will probably want to set
    # the source field manually

    class Meta:
        model = Placeholder


def get_plugin_position(plugin):
    """Helper function to correctly calculate the plugin position.
    Use this in plugin factory classes
    """
    offset = plugin.placeholder.get_last_plugin_position(plugin.language) or 0
    return offset + 1


def get_plugin_language(plugin):
    """Helper function to get the language from a plugin's relationships.
    Use this in plugin factory classes
    """
    if plugin.placeholder.source:
        return plugin.placeholder.source.language
    # NOTE: If plugin.placeholder.source is None then language will
    # also be None unless set manually


class TextPluginFactory(factory.django.DjangoModelFactory):
    language = factory.LazyAttribute(get_plugin_language)
    placeholder = factory.SubFactory(PlaceholderFactory)
    parent = None
    position = factory.LazyAttribute(get_plugin_position)
    plugin_type = "TextPlugin"
    body = factory.fuzzy.FuzzyText(length=50)

    class Meta:
        model = Text


class FancyPollFactory(factory.django.DjangoModelFactory):
    name = FuzzyText(length=24)

    class Meta:
        model = FancyPoll


class PollTitleExtensionFactory(factory.django.DjangoModelFactory):
    extended_object = factory.SubFactory(PageContentFactory)
    votes = FuzzyInteger(0, 100)

    class Meta:
        model = PollPageContentExtension


class TestTitleExtensionFactory(factory.django.DjangoModelFactory):
    extended_object = factory.SubFactory(PageContentFactory)

    class Meta:
        model = TestPageContentExtension
